"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
require("mocha");
const chai_1 = require("chai");
const C = require("../constants");
const http_authentication_handler_1 = require("./http-authentication-handler");
const test_http_server_1 = require("../test/helper/test-http-server");
const logger_mock_1 = require("../test/mock/logger-mock");
describe('it forwards authentication attempts as http post requests to a specified endpoint', () => {
    let authenticationHandler;
    let server;
    const port = test_http_server_1.default.getRandomPort();
    const logger = new logger_mock_1.default();
    before((done) => {
        server = new test_http_server_1.default(port, done);
    });
    after((done) => {
        server.close(done);
    });
    it('creates the authentication handler', () => {
        const endpointUrl = `http://localhost:${port}`;
        authenticationHandler = new http_authentication_handler_1.default({
            endpointUrl,
            permittedStatusCodes: [200],
            requestTimeout: 60,
        }, logger);
        chai_1.expect(authenticationHandler.description).to.equal(`http webhook to ${endpointUrl}`);
    });
    it('issues a request when isValidUser is called and receives 200 in return', (done) => {
        const connectionData = { connection: 'data' };
        const authData = { username: 'userA' };
        server.once('request-received', () => {
            chai_1.expect(server.lastRequestData).to.deep.equal({
                connectionData: { connection: 'data' },
                authData: { username: 'userA' }
            });
            chai_1.expect(server.lastRequestMethod).to.equal('POST');
            chai_1.expect(server.lastRequestHeaders['content-type']).to.contain('application/json');
            server.respondWith(200, { authData: { extra: 'data' } });
        });
        authenticationHandler.isValidUser(connectionData, authData, (result, data) => {
            chai_1.expect(result).to.equal(true);
            chai_1.expect(data).to.deep.equal({ authData: { extra: 'data' } });
            done();
        });
    });
    it('issues a request when isValidUser is called and receives 401 (denied) in return', (done) => {
        const connectionData = { connection: 'data' };
        const authData = { username: 'userA' };
        server.once('request-received', () => {
            chai_1.expect(server.lastRequestData).to.deep.equal({
                connectionData: { connection: 'data' },
                authData: { username: 'userA' }
            });
            chai_1.expect(server.lastRequestMethod).to.equal('POST');
            chai_1.expect(server.lastRequestHeaders['content-type']).to.contain('application/json');
            server.respondWith(401);
        });
        authenticationHandler.isValidUser(connectionData, authData, (result, data) => {
            chai_1.expect(result).to.equal(false);
            chai_1.expect(data).to.equal(null);
            chai_1.expect(logger._log).to.have.callCount(0);
            done();
        });
    });
    it('receives a positive response without data', (done) => {
        const connectionData = { connection: 'data' };
        const authData = { username: 'userA' };
        server.once('request-received', () => {
            chai_1.expect(server.lastRequestData).to.deep.equal({
                connectionData: { connection: 'data' },
                authData: { username: 'userA' }
            });
            chai_1.expect(server.lastRequestMethod).to.equal('POST');
            chai_1.expect(server.lastRequestHeaders['content-type']).to.contain('application/json');
            server.respondWith(200, '');
        });
        authenticationHandler.isValidUser(connectionData, authData, (result, data) => {
            chai_1.expect(result).to.equal(true);
            chai_1.expect(data).to.equal(null);
            done();
        });
    });
    it('receives a positive response with only a string', (done) => {
        const connectionData = { connection: 'data' };
        const authData = { username: 'userA' };
        server.once('request-received', () => {
            chai_1.expect(server.lastRequestData).to.deep.equal({
                connectionData: { connection: 'data' },
                authData: { username: 'userA' }
            });
            chai_1.expect(server.lastRequestMethod).to.equal('POST');
            chai_1.expect(server.lastRequestHeaders['content-type']).to.contain('application/json');
            server.respondWith(200, 'userA');
        });
        authenticationHandler.isValidUser(connectionData, authData, (result, data) => {
            chai_1.expect(result).to.equal(true);
            chai_1.expect(data).to.deep.equal({ username: 'userA' });
            done();
        });
    });
    it('receives a server error as response', (done) => {
        const connectionData = { connection: 'data' };
        const authData = { username: 'userA' };
        server.once('request-received', () => {
            server.respondWith(500, 'oh dear');
        });
        authenticationHandler.isValidUser(connectionData, authData, (result, data) => {
            chai_1.expect(result).to.equal(false);
            chai_1.expect(logger._log).to.have.been.calledWith(2, C.EVENT.AUTH_ERROR, 'http auth server error: "oh dear"');
            chai_1.expect(data).to.equal('oh dear');
            done();
        });
    });
    it.skip('times out', (done) => {
        const connectionData = { connection: 'data' };
        const authData = { username: 'userA' };
        server.once('request-received', () => {
            // don't respond
        });
        logger._log.resetHistory();
        authenticationHandler.isValidUser(connectionData, authData, (result, data) => {
            chai_1.expect(result).to.equal(false);
            chai_1.expect(logger._log).to.have.been.calledWith(2, C.EVENT.AUTH_ERROR, 'http auth error: Error: socket hang up');
            chai_1.expect(data).to.equalNull();
            server.respondWith(200);
            done();
        });
    });
});
//# sourceMappingURL=http-authentication-handler.spec.js.map