"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
require("mocha");
const chai_1 = require("chai");
const listener_registry_1 = require("./listener-registry");
const testHelper = require("../test/helper/test-helper");
const C = require("../constants");
const test_mocks_1 = require("../test/helper/test-mocks");
const sinon = require("sinon");
const options = testHelper.getDeepstreamOptions();
const config = options.config;
const services = options.services;
let topic;
let subscribedTopics;
const subscribers = new Set();
let clientRegistryMock;
let providers;
let clients;
let listenerRegistry;
let clientRegistry;
class ListenerTestUtils {
    constructor(listenerTopic) {
        topic = listenerTopic || C.TOPIC.RECORD;
        if (topic === C.TOPIC.RECORD) {
            this.actions = C.RECORD_ACTIONS;
        }
        else {
            this.actions = C.EVENT_ACTIONS;
        }
        subscribedTopics = [];
        clientRegistry = {
            hasName(subscriptionName) {
                return subscribedTopics.indexOf(subscriptionName);
            },
            getNames() {
                return subscribedTopics;
            },
            getLocalSubscribers() {
                return subscribers;
            },
            hasLocalSubscribers() {
                return subscribers.size > 0;
            },
            sendToSubscribers: () => { }
        };
        clientRegistryMock = sinon.mock(clientRegistry);
        config.listenResponseTimeout = 30;
        config.shuffleListenProviders = false;
        config.stateReconciliationTimeout = 10;
        clients = [
            null,
            test_mocks_1.getTestMocks().getSocketWrapper('c1'),
            test_mocks_1.getTestMocks().getSocketWrapper('c2'),
            test_mocks_1.getTestMocks().getSocketWrapper('c3')
        ];
        providers = [
            null,
            test_mocks_1.getTestMocks().getSocketWrapper('p1'),
            test_mocks_1.getTestMocks().getSocketWrapper('p2'),
            test_mocks_1.getTestMocks().getSocketWrapper('p3')
        ];
        listenerRegistry = new listener_registry_1.default(topic, config, services, clientRegistry);
        chai_1.expect(typeof listenerRegistry.handle).to.equal('function');
    }
    complete() {
        clients.forEach((client) => {
            if (client) {
                client.socketWrapperMock.verify();
            }
        });
        providers.forEach((provider) => {
            if (provider) {
                provider.socketWrapperMock.verify();
            }
        });
        clientRegistryMock.verify();
    }
    /**
    * Provider Utils
    */
    providerListensTo(provider, pattern) {
        providers[provider].socketWrapperMock
            .expects('sendAckMessage')
            .once()
            .withExactArgs({
            topic,
            action: this.actions.LISTEN,
            name: pattern
        });
        listenerRegistry.handle(providers[provider].socketWrapper, {
            topic,
            action: this.actions.LISTEN,
            name: pattern
        });
    }
    providerUnlistensTo(provider, pattern) {
        providers[provider].socketWrapperMock
            .expects('sendAckMessage')
            .once()
            .withExactArgs({
            topic,
            action: this.actions.UNLISTEN,
            name: pattern
        });
        listenerRegistry.handle(providers[provider].socketWrapper, {
            topic,
            action: this.actions.UNLISTEN,
            name: pattern
        });
    }
    providerWillGetSubscriptionFound(provider, pattern, subscription) {
        providers[provider].socketWrapperMock
            .expects('sendMessage')
            .once()
            .withExactArgs({
            topic,
            action: this.actions.SUBSCRIPTION_FOR_PATTERN_FOUND,
            name: pattern,
            subscription
        });
    }
    providerWillGetSubscriptionRemoved(provider, pattern, subscription) {
        providers[provider].socketWrapperMock
            .expects('sendMessage')
            .once()
            .withExactArgs({
            topic,
            action: this.actions.SUBSCRIPTION_FOR_PATTERN_REMOVED,
            name: pattern,
            subscription
        });
    }
    providerAcceptsButIsntAcknowledged(provider, pattern, subscriptionName) {
        this.providerAccepts(provider, pattern, subscriptionName, true);
    }
    providerAccepts(provider, pattern, subscription, doesnthaveActiveProvider) {
        listenerRegistry.handle(providers[provider].socketWrapper, {
            topic,
            action: this.actions.LISTEN_ACCEPT,
            name: pattern,
            subscription
        });
        chai_1.expect(listenerRegistry.hasActiveProvider(subscription)).to.equal(!doesnthaveActiveProvider);
    }
    providerRejectsAndPreviousTimeoutProviderThatAcceptedIsUsed(provider, pattern, subscriptionName) {
        this.providerRejects(provider, pattern, subscriptionName, true);
    }
    providerAcceptsAndIsSentSubscriptionRemoved(provider, pattern, subscriptionName) {
        this.providerWillGetSubscriptionRemoved(provider, pattern, subscriptionName);
        this.providerAcceptsButIsntAcknowledged(provider, pattern, subscriptionName);
    }
    providerRejects(provider, pattern, subscription, doNotCheckActiveProvider) {
        listenerRegistry.handle(providers[provider].socketWrapper, {
            topic,
            action: this.actions.LISTEN_REJECT,
            name: pattern,
            subscription
        });
        if (!doNotCheckActiveProvider) {
            chai_1.expect(listenerRegistry.hasActiveProvider(subscription)).to.equal(false);
        }
    }
    acceptMessageThrowsError(provider, pattern, subscription) {
        listenerRegistry.handle(providers[provider].socketWrapper, {
            topic,
            action: this.actions.LISTEN_ACCEPT,
            name: pattern,
            subscription
        });
        // TODO
        // verify( providers[ provider], this.actions.ERROR, [ C.EVENT.INVALID_MESSAGE, this.actions.LISTEN_ACCEPT, pattern, subscriptionName ] );
    }
    rejectMessageThrowsError(provider, pattern, subscription) {
        listenerRegistry.handle(providers[provider].socketWrapper, {
            topic,
            action: this.actions.LISTEN_REJECT,
            name: pattern,
            subscription
        });
        // TODO
        // verify( providers[ provider], this.actions.ERROR, [ C.EVENT.INVALID_MESSAGE, this.actions.LISTEN_REJECT, pattern, subscriptionName ] );
    }
    providerLosesItsConnection(provider) {
        providers[provider].socketWrapper.emit('close', providers[provider].socketWrapper);
    }
    /**
    * Subscriber Utils
    */
    subscriptionAlreadyMadeFor(subscriptionName) {
        subscribedTopics.push(subscriptionName);
    }
    clientSubscribesTo(client, subscriptionName, firstSubscription) {
        if (firstSubscription) {
            listenerRegistry.onFirstSubscriptionMade(subscriptionName);
        }
        listenerRegistry.onSubscriptionMade(subscriptionName, clients[client].socketWrapper);
        subscribedTopics.push(subscriptionName);
        subscribers.add(clients[client].socketWrapper);
    }
    clientUnsubscribesTo(client, subscriptionName, lastSubscription) {
        if (lastSubscription) {
            listenerRegistry.onLastSubscriptionRemoved(subscriptionName);
        }
        listenerRegistry.onSubscriptionRemoved(subscriptionName, clients[client].socketWrapper);
        subscribedTopics.splice(subscribedTopics.indexOf(subscriptionName), 1);
        subscribers.delete(clients[client].socketWrapper);
    }
    clientWillRecievePublishedUpdate(client, subscription, state) {
        clients[client].socketWrapperMock
            .expects('sendMessage')
            .once()
            .withExactArgs({
            topic,
            action: state ? this.actions.SUBSCRIPTION_HAS_PROVIDER : this.actions.SUBSCRIPTION_HAS_NO_PROVIDER,
            name: subscription,
        });
    }
    publishUpdateWillBeSentToSubscribers(subscription, state) {
        clientRegistryMock
            .expects('sendToSubscribers')
            .once()
            .withExactArgs(subscription, {
            topic,
            action: state ? this.actions.SUBSCRIPTION_HAS_PROVIDER : this.actions.SUBSCRIPTION_HAS_NO_PROVIDER,
            name: subscription,
        }, false, null);
    }
    subscriptionHasActiveProvider(subscription, value) {
        chai_1.expect(listenerRegistry.hasActiveProvider(subscription)).to.equal(value);
    }
}
exports.default = ListenerTestUtils;
//# sourceMappingURL=listener-test-utils.js.map