"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const connection_endpoint_1 = require("../websocket/connection-endpoint");
const constants_1 = require("../../constants");
const fileUtils = require("../../config/file-utils");
const binaryMessageBuilder = require("../../../binary-protocol/src/message-builder");
const binaryMessageParser = require("../../../binary-protocol/src/message-parser");
const socket_wrapper_factory_1 = require("./socket-wrapper-factory");
/**
 * This is the frontmost class of deepstream's message pipeline. It receives
 * connections and authentication requests, authenticates sockets and
 * forwards messages it receives from authenticated sockets.
 */
class UWSConnectionEndpoint extends connection_endpoint_1.default {
    constructor(options, services) {
        super(options, services);
        // alias require to trick nexe from bundling it
        const req = require;
        try {
            this.uWS = req('uWebSockets.js');
        }
        catch (e) {
            this.uWS = req(fileUtils.lookupLibRequirePath('uWebSockets.js'));
        }
        this.description = 'µWebSocket Connection Endpoint';
        this.onMessages = this.onMessages.bind(this);
        this.listenSocket = null;
    }
    /**
     * Initialize the uws endpoint, setup callbacks etc.
     *
     * @private
     * @returns {void}
     */
    createWebsocketServer() {
        this.connections = new Map();
        const options = {
            noDelay: this.getOption('noDelay'),
            perMessageDeflate: this.getOption('perMessageDeflate'),
            maxPayload: this.getOption('maxMessageSize')
        };
        const server = UWSConnectionEndpoint.getServer(this.uWS, {
            sslCert: this.getOption('sslCert'),
            sslKey: this.getOption('sslKey'),
            sslDHParams: this.getOption('sslDHParams'),
            passphrase: this.getOption('sslPassphrase')
        }, options);
        server.get(this.getOption('healthCheckPath'), (res) => {
            res.end();
        });
        server.ws(this.getOption('urlPath'), {
            /* Options */
            compression: 0,
            maxPayloadLength: this.getOption('maxMessageSize'),
            idleTimeout: this.getOption('heartbeatInterval') * 2,
            /* Handlers */
            open: (ws, request) => {
                const socketWrapper = this.createWebsocketWrapper(ws, request);
                this.connections.set(ws, socketWrapper);
                this.onConnection(socketWrapper);
            },
            message: (ws, message) => {
                this.connections.get(ws).onMessage(binaryMessageParser.parse(Buffer.from(message.slice())));
            },
            drain: () => {
            },
            close: (ws) => {
                this.onSocketClose(this.connections.get(ws));
                this.connections.delete(ws);
            }
        });
        server.listen(this.getOption('host'), this.getOption('port'), (token) => {
            /* Save the listen socket for later shut down */
            this.listenSocket = token;
            if (token) {
                this.onReady();
                const pingMessage = binaryMessageBuilder.getMessage({ topic: constants_1.TOPIC.CONNECTION, action: constants_1.CONNECTION_ACTIONS.PING }, false);
                this.pingInterval = setInterval(() => {
                    this.connections.forEach((con) => {
                        if (!con.isClosed) {
                            con.sendBinaryMessage(pingMessage);
                        }
                    });
                }, this.getOption('heartbeatInterval'));
            }
            else {
                this.logger.error(constants_1.STATES.SERVICE_INIT, `Failed to listen to port ${this.getOption('port')}`);
            }
        });
        return server;
    }
    /**
     * Returns sslKey, sslCert and other options from the config.
     *
     * @throws Will throw an error if one of sslKey or sslCert are not specified
     *
     * @private
     * @returns {null|Object} {
     *   {String}           key   ssl key
     *   {String}           cert  ssl certificate
     *   {String|undefined} ca    ssl certificate authority (if it's present in options)
     * }
     */
    static getSLLParams(config) {
        const keyFileName = config.sslKey;
        const certFileName = config.sslCert;
        const dhParamsFile = config.sslDHParams;
        const passphrase = config.sslPassphrase;
        if (keyFileName || certFileName) {
            if (!keyFileName) {
                throw new Error('Must also include sslKey in order to use SSL');
            }
            if (!certFileName) {
                throw new Error('Must also include sslCertFile in order to use SSL');
            }
            return {
                key_file_name: keyFileName,
                cert_file_name: certFileName,
                passphrase,
                dh_params_file_name: dhParamsFile
            };
        }
        return null;
    }
    static getServer(uWS, config, options) {
        let server;
        const sslParams = UWSConnectionEndpoint.getSLLParams(config);
        if (sslParams) {
            server = new uWS.SSLApp(Object.assign({}, options, sslParams));
        }
        else {
            server = new uWS.App(options);
        }
        return server;
    }
    static getHeaders(desiredHeaders, req) {
        const headers = {};
        for (const wantedHeader of desiredHeaders) {
            headers[wantedHeader] = req.getHeader(wantedHeader);
        }
        return headers;
    }
    closeWebsocketServer() {
        this.connections.forEach((conn) => {
            if (!conn.isClosed) {
                conn.socket.close();
            }
        });
        this.uWS.us_listen_socket_close(this.listenSocket);
        this.connections.clear();
        clearInterval(this.pingInterval);
        setTimeout(() => this.emit('close'), 2000);
    }
    /**
     * Receives a connected socket, wraps it in a SocketWrapper, sends a connection ack to the user
     * and subscribes to authentication messages.
     * @param {Websocket} socket
     *
     * @param {WebSocket} external    uws native websocket
     *
     * @private
     * @returns {void}
     */
    createWebsocketWrapper(websocket, upgradeReq) {
        const handshakeData = {
            remoteAddress: new Uint8Array(websocket.getRemoteAddress()).join('.'),
            headers: UWSConnectionEndpoint.getHeaders(this.options.headers, upgradeReq),
            referer: upgradeReq.getHeader('referer')
        };
        const socketWrapper = socket_wrapper_factory_1.createUWSSocketWrapper(websocket, handshakeData, this.logger, this.options, this);
        return socketWrapper;
    }
    onSocketWrapperClosed(socketWrapper) {
        socketWrapper.close();
    }
}
exports.default = UWSConnectionEndpoint;
//# sourceMappingURL=connection-endpoint.js.map