"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const constants_1 = require("../../constants");
const binaryMessageBuilder = require("../../../binary-protocol/src/message-builder");
const binaryMessageParser = require("../../../binary-protocol/src/message-parser");
/**
 * This class wraps around a websocket
 * and provides higher level methods that are integrated
 * with deepstream's message structure
 */
class UwsSocketWrapper {
    constructor(socket, handshakeData, logger, config, connectionEndpoint) {
        this.socket = socket;
        this.handshakeData = handshakeData;
        this.logger = logger;
        this.config = config;
        this.connectionEndpoint = connectionEndpoint;
        this.isRemote = false;
        this.isClosed = false;
        this.uuid = Math.random();
        this.authAttempts = 0;
        this.closeCallbacks = new Set();
    }
    get isOpen() {
        return this.isClosed !== true;
    }
    /**
     * Called by the connection endpoint to flush all buffered writes.
     * A buffered write is a write that is not a high priority, such as an ack
     * and can wait to be bundled into another message if necessary
     */
    flush() {
        if (this.bufferedWrites.length !== 0) {
            this.socket.send(Buffer.concat(this.bufferedWrites), true);
            this.bufferedWrites = [];
        }
    }
    /**
     * Sends a message based on the provided action and topic
     * @param {Boolean} allowBuffering Boolean to indicate that buffering is allowed on
     *                                 this message type
     */
    sendMessage(message, allowBuffering) {
        this.sendBinaryMessage(binaryMessageBuilder.getMessage(message, false), allowBuffering);
    }
    /**
     * Sends a message based on the provided action and topic
     * @param {Boolean} allowBuffering Boolean to indicate that buffering is allowed on
     *                                 this message type
     */
    sendAckMessage(message, allowBuffering) {
        this.sendBinaryMessage(binaryMessageBuilder.getMessage(message, true), true);
    }
    getMessage(message) {
        return binaryMessageBuilder.getMessage(message, false);
    }
    parseMessage(message) {
        /* we copy the underlying buffer (since a shallow reference won't be safe
         * outside of the callback)
         * the copy could be avoided if we make sure not to store references to the
         * raw buffer within the message
         */
        return binaryMessageParser.parse(Buffer.from(Buffer.from(message)));
    }
    parseData(message) {
        return binaryMessageParser.parseData(message);
    }
    onMessage(messages) {
    }
    /**
     * Destroys the socket. Removes all deepstream specific
     * logic and closes the connection
     */
    destroy() {
        this.socket.end();
    }
    close() {
        this.isClosed = true;
        delete this.authCallback;
        this.closeCallbacks.forEach((cb) => cb(this));
        this.logger.info(constants_1.EVENT.CLIENT_DISCONNECTED, this.user);
    }
    /**
     * Returns a map of parameters that were collected
     * during the initial http request that established the
     * connection
     */
    getHandshakeData() {
        return this.handshakeData;
    }
    onClose(callback) {
        this.closeCallbacks.add(callback);
    }
    removeOnClose(callback) {
        this.closeCallbacks.delete(callback);
    }
    sendBinaryMessage(binaryMessage, allowBuffering) {
        if (this.isOpen) {
            if (this.config.outgoingBufferTimeout === 0) {
                this.socket.send(binaryMessage, true);
            }
            else if (!allowBuffering) {
                this.flush();
                this.socket.send(Buffer.concat(this.bufferedWrites), true);
                this.bufferedWrites = [];
            }
            else {
                this.bufferedWrites.push(binaryMessage);
                this.connectionEndpoint.scheduleFlush(this);
            }
        }
    }
}
exports.UwsSocketWrapper = UwsSocketWrapper;
exports.createUWSSocketWrapper = function (socket, handshakeData, logger, config, connectionEndpoint) { return new UwsSocketWrapper(socket, handshakeData, logger, config, connectionEndpoint); };
//# sourceMappingURL=socket-wrapper-factory.js.map