"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const events_1 = require("events");
const message_constants_1 = require("../../../binary-protocol/src/message-constants");
const OPEN = 'OPEN';
var ClientEvent;
(function (ClientEvent) {
    ClientEvent["CLIENT_CONNECTED"] = "client-connected";
    ClientEvent["CLIENT_DISCONNECTED"] = "client-disconnected";
})(ClientEvent || (ClientEvent = {}));
/**
 * This is the frontmost class of deepstream's message pipeline. It receives
 * connections and authentication requests, authenticates sockets and
 * forwards messages it receives from authenticated sockets.
 */
class WebsocketConnectionEndpoint extends events_1.EventEmitter {
    constructor(options, services) {
        super();
        this.options = options;
        this.services = services;
        this.isReady = false;
        this.description = 'µWebSocket Connection Endpoint';
        this.initialised = false;
        this.authenticatedSocketWrappers = new Set();
        this.scheduledSocketWrapperWrites = new Set();
        this.flushSockets = this.flushSockets.bind(this);
    }
    createWebsocketServer() {
    }
    closeWebsocketServer() {
    }
    onSocketWrapperClosed(socketWrapper) {
    }
    /**
     * Called for every message that's received
     * from an authenticated socket
     *
     * This method will be overridden by an external class and is used instead
     * of an event emitter to improve the performance of the messaging pipeline
     */
    onMessages(socketWrapper, messages) {
    }
    /**
     * Called on initialization with a reference to the instantiating deepstream server.
     */
    setDeepstream(deepstream) {
        this.dsOptions = deepstream.config;
        this.logger = deepstream.services.logger;
        this.authenticationHandler = deepstream.services.authenticationHandler;
    }
    /**
     * Initialise and setup the http and WebSocket servers.
     *
     * @throws Will throw if called before `setDeepstream()`.
     */
    init() {
        if (!this.dsOptions) {
            throw new Error('setDeepstream must be called before init()');
        }
        if (this.initialised) {
            throw new Error('init() must only be called once');
        }
        this.initialised = true;
        this.maxAuthAttempts = this.getOption('maxAuthAttempts');
        this.logInvalidAuthData = this.getOption('logInvalidAuthData');
        this.urlPath = this.getOption('urlPath');
        this.unauthenticatedClientTimeout = this.getOption('unauthenticatedClientTimeout');
        this.websocketServer = this.createWebsocketServer();
    }
    /**
     * Called from a socketWrapper. This method tells the connection endpoint
     * to flush the socket after a certain amount of time, used to low priority
     * messages
     */
    scheduleFlush(socketWrapper) {
        this.scheduledSocketWrapperWrites.add(socketWrapper);
        if (!this.flushTimeout) {
            this.flushTimeout = setTimeout(this.flushSockets, this.options.outgoingBufferTimeout);
        }
    }
    /**
     * Called when the flushTimeout occurs in order to send  all pending socket acks
     */
    flushSockets() {
        for (const socketWrapper of this.scheduledSocketWrapperWrites) {
            socketWrapper.flush();
        }
        this.scheduledSocketWrapperWrites.clear();
        this.flushTimeout = null;
    }
    /**
     * Get a parameter from the root of the deepstream options if present, otherwise get it from the
     * plugin config. If neither is present, default to the optionally provided default.
     */
    getOption(option) {
        const value = this.dsOptions[option];
        if ((value === null || value === undefined) && (this.options[option] !== undefined)) {
            return this.options[option];
        }
        return value;
    }
    handleParseErrors(socketWrapper, parseResults) {
        const messages = [];
        for (const parseResult of parseResults) {
            if (parseResult.parseError) {
                const rawMsg = this.getRaw(parseResult);
                this.logger.warn(message_constants_1.PARSER_ACTIONS[message_constants_1.PARSER_ACTIONS.MESSAGE_PARSE_ERROR], `error parsing connection message ${rawMsg}`);
                socketWrapper.sendMessage({
                    topic: message_constants_1.TOPIC.PARSER,
                    action: parseResult.action,
                    data: parseResult.raw,
                    originalTopic: parseResult.parsedMessage.topic,
                    originalAction: parseResult.parsedMessage.action
                });
                socketWrapper.destroy();
                continue;
            }
            const message = parseResult;
            if (message.topic === message_constants_1.TOPIC.CONNECTION &&
                message.action === message_constants_1.CONNECTION_ACTIONS.PONG) {
                continue;
            }
            messages.push(message);
        }
        return messages;
    }
    getRaw(parseResult) {
        if (parseResult.raw && typeof parseResult.raw === 'string') {
            return parseResult.raw;
        }
        else if (parseResult.parseError && parseResult.parsedMessage) {
            return JSON.stringify(parseResult.parsedMessage);
        }
        else if (parseResult.raw instanceof Buffer) {
            return JSON.stringify(parseResult.raw);
        }
        return '';
    }
    /**
     * Called for the ready event of the ws server.
     */
    onReady() {
        const wsMsg = `Listening for websocket connections on ${this.getOption('host')}:${this.getOption('port')}${this.urlPath}`;
        this.logger.info(message_constants_1.EVENT.INFO, wsMsg);
        const hcMsg = `Listening for health checks on path ${this.getOption('healthCheckPath')} `;
        this.logger.info(message_constants_1.EVENT.INFO, hcMsg);
        this.emit('ready');
        this.isReady = true;
    }
    /**
     * Receives a connected socket, wraps it in a SocketWrapper, sends a connection ack to the user
     * and subscribes to authentication messages.
     */
    onConnection(socketWrapper) {
        const handshakeData = socketWrapper.getHandshakeData();
        this.logger.info(message_constants_1.EVENT.INCOMING_CONNECTION, `from ${handshakeData.referer} (${handshakeData.remoteAddress})`);
        let disconnectTimer;
        if (this.unauthenticatedClientTimeout !== null && this.unauthenticatedClientTimeout !== false) {
            const timeout = this.unauthenticatedClientTimeout;
            disconnectTimer = setTimeout(this.processConnectionTimeout.bind(this, socketWrapper), timeout);
            socketWrapper.onClose(clearTimeout.bind(null, disconnectTimer));
        }
        socketWrapper.authCallback = this.authenticateConnection.bind(this, socketWrapper, disconnectTimer);
        socketWrapper.onMessage = this.processConnectionMessage.bind(this, socketWrapper);
    }
    /**
     * Always challenges the client that connects. This will be opened up later to allow users
     * to put in their own challenge authentication.
     */
    processConnectionMessage(socketWrapper, parsedMessages) {
        const msg = parsedMessages[0];
        if (msg.topic !== message_constants_1.TOPIC.CONNECTION) {
            const rawMessage = this.getRaw(msg);
            this.logger.warn(message_constants_1.CONNECTION_ACTIONS[message_constants_1.CONNECTION_ACTIONS.INVALID_MESSAGE], `invalid connection message ${rawMessage}`);
            socketWrapper.sendMessage({
                topic: message_constants_1.TOPIC.CONNECTION,
                action: message_constants_1.CONNECTION_ACTIONS.INVALID_MESSAGE,
                originalTopic: msg.topic,
                originalAction: msg.action,
                data: rawMessage
            });
            return;
        }
        if (msg.action === message_constants_1.CONNECTION_ACTIONS.CHALLENGE) {
            socketWrapper.onMessage = socketWrapper.authCallback;
            socketWrapper.sendMessage({
                topic: message_constants_1.TOPIC.CONNECTION,
                action: message_constants_1.CONNECTION_ACTIONS.ACCEPT
            });
            return;
        }
        this.logger.error(message_constants_1.PARSER_ACTIONS[message_constants_1.PARSER_ACTIONS.UNKNOWN_ACTION], '', msg.action);
    }
    /**
     * Callback for the first message that's received from the socket.
     * This is expected to be an auth-message. This method makes sure that's
     * the case and - if so - forwards it to the permission handler for authentication
     */
    authenticateConnection(socketWrapper, disconnectTimeout, parsedMessages) {
        const msg = parsedMessages[0];
        let errorMsg;
        if (msg.topic !== message_constants_1.TOPIC.AUTH) {
            const rawMessage = this.getRaw(msg);
            this.logger.warn(message_constants_1.AUTH_ACTIONS[message_constants_1.AUTH_ACTIONS.INVALID_MESSAGE], `invalid auth message ${rawMessage}`);
            socketWrapper.sendMessage({
                topic: message_constants_1.TOPIC.AUTH,
                action: message_constants_1.AUTH_ACTIONS.INVALID_MESSAGE,
                originalTopic: msg.topic,
                originalAction: msg.action,
                data: rawMessage
            });
            return;
        }
        /**
         * Log the authentication attempt
         */
        const logMsg = socketWrapper.getHandshakeData().remoteAddress;
        this.logger.debug(message_constants_1.AUTH_ACTIONS[message_constants_1.AUTH_ACTIONS.REQUEST], logMsg);
        /**
         * Ensure the message is a valid authentication message
         */
        if (msg.action !== message_constants_1.AUTH_ACTIONS.REQUEST) {
            errorMsg = this.logInvalidAuthData === true ? msg.data : '';
            this.sendInvalidAuthMsg(socketWrapper, errorMsg, msg.action);
            return;
        }
        /**
         * Ensure the authentication data is valid JSON
         */
        const result = socketWrapper.parseData(msg);
        if (result instanceof Error || !msg.parsedData || typeof msg.parsedData !== 'object') {
            errorMsg = 'Error parsing auth message';
            if (this.logInvalidAuthData === true) {
                errorMsg += ` "${msg.data}": ${result.toString()}`;
            }
            this.sendInvalidAuthMsg(socketWrapper, errorMsg, msg.action);
            return;
        }
        /**
         * Forward for authentication
         */
        this.authenticationHandler.isValidUser(socketWrapper.getHandshakeData(), msg.parsedData, this.processAuthResult.bind(this, msg.parsedData, socketWrapper, disconnectTimeout));
    }
    /**
     * Will be called for syntactically incorrect auth messages. Logs
     * the message, sends an error to the client and closes the socket
     */
    sendInvalidAuthMsg(socketWrapper, msg, originalAction) {
        this.logger.warn(message_constants_1.AUTH_ACTIONS[message_constants_1.AUTH_ACTIONS.INVALID_MESSAGE_DATA], this.logInvalidAuthData ? msg : '');
        socketWrapper.sendMessage({
            topic: message_constants_1.TOPIC.AUTH,
            action: message_constants_1.AUTH_ACTIONS.INVALID_MESSAGE_DATA,
            originalAction
        });
        socketWrapper.destroy();
    }
    /**
     * Callback for succesfully validated sockets. Removes
     * all authentication specific logic and registeres the
     * socket with the authenticated sockets
     */
    registerAuthenticatedSocket(socketWrapper, userData) {
        this.authenticatedSocketWrappers.add(socketWrapper);
        delete socketWrapper.authCallback;
        socketWrapper.onMessage = (parsedMessages) => {
            this.onMessages(socketWrapper, parsedMessages);
        };
        this.appendDataToSocketWrapper(socketWrapper, userData);
        socketWrapper.sendMessage({
            topic: message_constants_1.TOPIC.AUTH,
            action: message_constants_1.AUTH_ACTIONS.AUTH_SUCCESSFUL,
            parsedData: userData.clientData
        });
        if (socketWrapper.user !== OPEN) {
            this.emit(ClientEvent.CLIENT_CONNECTED, socketWrapper);
        }
        this.logger.info(message_constants_1.AUTH_ACTIONS[message_constants_1.AUTH_ACTIONS.AUTH_SUCCESSFUL], socketWrapper.user);
    }
    /**
     * Append connection data to the socket wrapper
     */
    appendDataToSocketWrapper(socketWrapper, userData) {
        socketWrapper.user = userData.username || OPEN;
        socketWrapper.authData = userData.serverData || null;
        socketWrapper.clientData = userData.clientData || null;
    }
    /**
     * Callback for invalid credentials. Will notify the client
     * of the invalid auth attempt. If the number of invalid attempts
     * exceed the threshold specified in options.maxAuthAttempts
     * the client will be notified and the socket destroyed.
     */
    processInvalidAuth(clientData, authData, socketWrapper) {
        let logMsg = 'invalid authentication data';
        if (this.logInvalidAuthData === true) {
            logMsg += `: ${JSON.stringify(authData)}`;
        }
        this.logger.info(message_constants_1.AUTH_ACTIONS[message_constants_1.AUTH_ACTIONS.AUTH_UNSUCCESSFUL], logMsg);
        socketWrapper.sendMessage({
            topic: message_constants_1.TOPIC.AUTH,
            action: message_constants_1.AUTH_ACTIONS.AUTH_UNSUCCESSFUL,
            parsedData: clientData
        });
        socketWrapper.authAttempts++;
        if (socketWrapper.authAttempts >= this.maxAuthAttempts) {
            this.logger.info(message_constants_1.AUTH_ACTIONS[message_constants_1.AUTH_ACTIONS.TOO_MANY_AUTH_ATTEMPTS], 'too many authentication attempts');
            socketWrapper.sendMessage({
                topic: message_constants_1.TOPIC.AUTH,
                action: message_constants_1.AUTH_ACTIONS.TOO_MANY_AUTH_ATTEMPTS
            });
            socketWrapper.destroy();
        }
    }
    /**
     * Callback for connections that have not authenticated succesfully within
     * the expected timeframe
     */
    processConnectionTimeout(socketWrapper) {
        const log = 'connection has not authenticated successfully in the expected time';
        this.logger.info(message_constants_1.CONNECTION_ACTIONS[message_constants_1.CONNECTION_ACTIONS.AUTHENTICATION_TIMEOUT], log);
        socketWrapper.sendMessage({
            topic: message_constants_1.TOPIC.CONNECTION,
            action: message_constants_1.CONNECTION_ACTIONS.AUTHENTICATION_TIMEOUT
        });
        socketWrapper.destroy();
    }
    /**
     * Callback for the results returned by the permissionHandler
     */
    processAuthResult(authData, socketWrapper, disconnectTimeout, isAllowed, userData) {
        userData = userData || {};
        clearTimeout(disconnectTimeout);
        if (isAllowed === true) {
            this.registerAuthenticatedSocket(socketWrapper, userData);
        }
        else {
            this.processInvalidAuth(userData.clientData, authData, socketWrapper); // todo
        }
    }
    /**
     * Notifies the (optional) onClientDisconnect method of the permissionHandler
     * that the specified client has disconnected
     */
    onSocketClose(socketWrapper) {
        this.scheduledSocketWrapperWrites.delete(socketWrapper);
        this.onSocketWrapperClosed(socketWrapper);
        if (this.authenticatedSocketWrappers.delete(socketWrapper)) {
            if (this.authenticationHandler.onClientDisconnect) {
                this.authenticationHandler.onClientDisconnect(socketWrapper.user);
            }
            if (socketWrapper.user !== OPEN) {
                this.emit(ClientEvent.CLIENT_DISCONNECTED, socketWrapper);
            }
        }
    }
    /**
     * Closes the ws server connection. The ConnectionEndpoint
     * will emit a close event once succesfully shut down
     */
    close() {
        this.closeWebsocketServer();
    }
}
exports.default = WebsocketConnectionEndpoint;
//# sourceMappingURL=connection-endpoint.js.map