"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
require("mocha");
const chai_1 = require("chai");
const pathParser = require("./path-parser");
const isRegExp = function (val) {
    return typeof val === 'object' && typeof val.test === 'function';
};
describe('validates paths in permission.json files', () => {
    it('exposes a validate method', () => {
        chai_1.expect(typeof pathParser.validate).to.equal('function');
    });
    it('accepts a valid path', () => {
        chai_1.expect(pathParser.validate('game-comment/$gameId/*')).to.equal(true);
    });
    it('rejects none strings', () => {
        chai_1.expect(pathParser.validate(3)).to.equal('path must be a string');
    });
    it('rejects empty strings', () => {
        chai_1.expect(pathParser.validate('')).to.equal('path can\'t be empty');
    });
    it('rejects paths starting with /', () => {
        chai_1.expect(pathParser.validate('/bla')).to.equal('path can\'t start with /');
    });
    it('rejects paths with invalid variable names', () => {
        chai_1.expect(pathParser.validate('bla/$-')).to.equal('invalid variable name $-');
        chai_1.expect(pathParser.validate('bla/$$aa')).to.equal('invalid variable name $$');
    });
});
describe('parses valid paths in permission.json files', () => {
    it('exposes a parse method', () => {
        chai_1.expect(typeof pathParser.parse).to.equal('function');
    });
    it('parses a simple, valid path', () => {
        const result = pathParser.parse('i-am-valid');
        chai_1.expect(isRegExp(result.regexp)).to.equal(true);
        chai_1.expect(result.regexp.toString()).to.equal('/^i-am-valid$/');
        chai_1.expect(result.variables.length).to.equal(0);
    });
    it('parses a valid path with a wildcard', () => {
        const result = pathParser.parse('i-am-valid/*');
        chai_1.expect(isRegExp(result.regexp)).to.equal(true);
        chai_1.expect(result.regexp.toString()).to.equal('/^i-am-valid\\/.*$/');
        chai_1.expect(result.variables.length).to.equal(0);
    });
    it('parses a valid path with a variable', () => {
        const result = pathParser.parse('game-score/$gameId');
        chai_1.expect(isRegExp(result.regexp)).to.equal(true);
        chai_1.expect(result.regexp.toString()).to.equal('/^game-score\\/([^\\/]+)$/');
        chai_1.expect(result.variables).to.deep.equal(['$gameId']);
    });
    it('parses a valid path with multiple variables', () => {
        const result = pathParser.parse('game-comment/$gameId/$userId/$commentId');
        chai_1.expect(isRegExp(result.regexp)).to.equal(true);
        chai_1.expect(result.regexp.toString()).to.equal('/^game-comment\\/([^\\/]+)\\/([^\\/]+)\\/([^\\/]+)$/');
        chai_1.expect(result.variables).to.deep.equal(['$gameId', '$userId', '$commentId']);
    });
    it('parses a path with a mix of variables and wildcards', () => {
        const result = pathParser.parse('$recordName/*');
        chai_1.expect(isRegExp(result.regexp)).to.equal(true);
        chai_1.expect(result.regexp.toString()).to.equal('/^([^\\/]+)\\/.*$/');
        chai_1.expect(result.variables).to.deep.equal(['$recordName']);
    });
});
describe('applies regexp to paths', () => {
    it('applies a regexp to a simple path', () => {
        const path = 'public/*';
        const result = pathParser.parse(path);
        chai_1.expect(result.regexp.test('public/details/info')).to.equal(true);
        chai_1.expect(result.regexp.test('private/details/info')).to.equal(false);
    });
    it('applies a regexp and extracts a variable from a simple path', () => {
        const path = 'private/$userId';
        const name = 'private/userA';
        const result = pathParser.parse(path);
        chai_1.expect(result.regexp.test(name)).to.equal(true);
        const r = name.match(result.regexp) || false;
        chai_1.expect(r[1]).to.equal('userA');
    });
    it('applies a regexp and extracts variables from a more complex path', () => {
        const path = 'private/$userId/*/$anotherId';
        const name = 'private/userA/blabla/14';
        const result = pathParser.parse(path);
        chai_1.expect(result.regexp.test(name)).to.equal(true);
        const r = name.match(result.regexp) || [];
        chai_1.expect(r.join(',')).to.deep.equal('private/userA/blabla/14,userA,14');
        const reject = name.match(result.regexp) || false;
        chai_1.expect(reject[1]).to.equal('userA');
    });
});
//# sourceMappingURL=path-parser.spec.js.map