"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
require("mocha");
const chai_1 = require("chai");
const sinon_1 = require("sinon");
const record_deletion_1 = require("../../src/record/record-deletion");
const M = require("./test-messages");
const C = require("../../src/constants");
const testHelper = require("../test/helper/test-helper");
const test_mocks_1 = require("../test/helper/test-mocks");
describe('record deletion', () => {
    let testMocks;
    let recordDeletion;
    let client;
    let config;
    let services;
    let callback;
    beforeEach(() => {
        testMocks = test_mocks_1.getTestMocks();
        client = testMocks.getSocketWrapper();
        const options = testHelper.getDeepstreamOptions();
        config = options.config;
        services = options.services;
        callback = sinon_1.spy();
    });
    afterEach(() => {
        client.socketWrapperMock.verify();
    });
    it('deletes records - happy path', () => {
        client.socketWrapperMock
            .expects('sendMessage')
            .once()
            .withExactArgs(M.deletionSuccessMsg);
        recordDeletion = new record_deletion_1.default(config, services, client.socketWrapper, M.deletionMsg, callback);
        chai_1.expect(services.cache.completedDeleteOperations).to.equal(1);
        chai_1.expect(services.storage.completedDeleteOperations).to.equal(1);
        chai_1.expect(recordDeletion.isDestroyed).to.equal(true);
        chai_1.expect(callback).to.have.callCount(1);
    });
    it('encounters an error during record deletion', (done) => {
        services.cache.nextOperationWillBeSuccessful = false;
        services.cache.nextOperationWillBeSynchronous = false;
        client.socketWrapperMock
            .expects('sendMessage')
            .once()
            .withExactArgs({
            topic: C.TOPIC.RECORD,
            action: C.RECORD_ACTIONS.RECORD_DELETE_ERROR,
            name: 'someRecord'
        });
        recordDeletion = new record_deletion_1.default(config, services, client.socketWrapper, M.deletionMsg, callback);
        setTimeout(() => {
            chai_1.expect(recordDeletion.isDestroyed).to.equal(true);
            chai_1.expect(callback).to.have.callCount(0);
            chai_1.expect(services.logger._log.firstCall.args).to.deep.equal([3, C.RECORD_ACTIONS[C.RECORD_ACTIONS.RECORD_DELETE_ERROR], 'storageError']);
            done();
        }, 20);
    });
    it('encounters an ack delete timeout', (done) => {
        config.cacheRetrievalTimeout = 10;
        services.cache.nextOperationWillBeSuccessful = false;
        services.cache.nextOperationWillBeSynchronous = false;
        client.socketWrapperMock
            .expects('sendMessage')
            .once()
            .withExactArgs({
            topic: C.TOPIC.RECORD,
            action: C.RECORD_ACTIONS.RECORD_DELETE_ERROR,
            name: 'someRecord'
        });
        recordDeletion = new record_deletion_1.default(config, services, client.socketWrapper, M.deletionMsg, callback);
        setTimeout(() => {
            chai_1.expect(recordDeletion.isDestroyed).to.equal(true);
            chai_1.expect(callback).to.have.callCount(0);
            chai_1.expect(services.logger._log.firstCall.args).to.deep.equal([3, C.RECORD_ACTIONS[C.RECORD_ACTIONS.RECORD_DELETE_ERROR], 'cache timeout']);
            done();
        }, 100);
    });
    it('doesn\'t delete excluded messages from storage', () => {
        config.storageExclusionPatterns = ['no-storage/'];
        client.socketWrapperMock
            .expects('sendMessage')
            .once()
            .withExactArgs(M.anotherDeletionSuccessMsg);
        recordDeletion = new record_deletion_1.default(config, services, client.socketWrapper, M.anotherDeletionMsg, callback);
        chai_1.expect(services.cache.completedDeleteOperations).to.equal(1);
        chai_1.expect(services.storage.completedDeleteOperations).to.equal(0);
        chai_1.expect(recordDeletion.isDestroyed).to.equal(true);
        chai_1.expect(callback).to.have.callCount(1);
    });
});
//# sourceMappingURL=record-deletion.spec.js.map