"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
require("mocha");
const chai_1 = require("chai");
const sinon_1 = require("sinon");
const C = require("../constants");
const recordRequest = require('./record-request').recordRequest;
const test_mocks_1 = require("../test/helper/test-mocks");
const testHelper = require('../test/helper/test-helper');
describe('record request', () => {
    const completeCallback = sinon_1.spy();
    const errorCallback = sinon_1.spy();
    let testMocks;
    let client;
    let config;
    let services;
    const cacheData = { cache: true };
    const storageData = { storage: true };
    beforeEach(() => {
        const options = testHelper.getDeepstreamOptions();
        services = options.services;
        config = Object.assign({}, options.config, {
            cacheRetrievalTimeout: 100,
            storageRetrievalTimeout: 100,
            storageExclusionPrefixes: ['dont-save']
        });
        services.cache.set('existingRecord', 1, cacheData, () => { });
        services.storage.set('onlyExistsInStorage', 1, storageData, () => { });
        testMocks = test_mocks_1.getTestMocks();
        client = testMocks.getSocketWrapper('someUser');
        completeCallback.resetHistory();
        errorCallback.resetHistory();
    });
    describe('records are requested from cache and storage sequentially', () => {
        it('requests a record that exists in a synchronous cache', () => {
            services.cache.nextOperationWillBeSynchronous = true;
            recordRequest('existingRecord', config, services, client.socketWrapper, completeCallback, errorCallback, null);
            chai_1.expect(services.cache.lastRequestedKey).to.equal('existingRecord');
            chai_1.expect(services.storage.lastRequestedKey).to.equal(null);
            chai_1.expect(completeCallback).to.have.been.calledWith('existingRecord', 1, cacheData, client.socketWrapper);
            chai_1.expect(errorCallback).to.have.callCount(0);
        });
        it('requests a record that exists in an asynchronous cache', (done) => {
            services.cache.nextGetWillBeSynchronous = false;
            recordRequest('existingRecord', config, services, client.socketWrapper, completeCallback, errorCallback, null);
            setTimeout(() => {
                chai_1.expect(completeCallback).to.have.been.calledWith('existingRecord', 1, cacheData, client.socketWrapper);
                chai_1.expect(errorCallback).to.have.callCount(0);
                chai_1.expect(services.cache.lastRequestedKey).to.equal('existingRecord');
                chai_1.expect(services.storage.lastRequestedKey).to.equal(null);
                done();
            }, 30);
        });
        it('requests a record that doesn\'t exists in a synchronous cache, but in storage', () => {
            services.cache.nextGetWillBeSynchronous = true;
            recordRequest('onlyExistsInStorage', config, services, client.socketWrapper, completeCallback, errorCallback, null);
            chai_1.expect(services.cache.lastRequestedKey).to.equal('onlyExistsInStorage');
            chai_1.expect(services.storage.lastRequestedKey).to.equal('onlyExistsInStorage');
            chai_1.expect(completeCallback).to.have.been.calledWith('onlyExistsInStorage', 1, storageData, client.socketWrapper);
            chai_1.expect(errorCallback).to.have.callCount(0);
        });
        it('requests a record that doesn\'t exists in an asynchronous cache, but in asynchronous storage', (done) => {
            services.cache.nextGetWillBeSynchronous = false;
            services.storage.nextGetWillBeSynchronous = false;
            recordRequest('onlyExistsInStorage', config, services, client.socketWrapper, completeCallback, errorCallback, null);
            setTimeout(() => {
                chai_1.expect(services.cache.lastRequestedKey).to.equal('onlyExistsInStorage');
                chai_1.expect(services.storage.lastRequestedKey).to.equal('onlyExistsInStorage');
                chai_1.expect(errorCallback).to.have.callCount(0);
                chai_1.expect(completeCallback).to.have.been.calledWith('onlyExistsInStorage', 1, storageData, client.socketWrapper);
                done();
            }, 75);
        });
        it('returns null for non existent records', () => {
            services.cache.nextGetWillBeSynchronous = true;
            recordRequest('doesNotExist', config, services, client.socketWrapper, completeCallback, errorCallback, null);
            chai_1.expect(completeCallback).to.have.been.calledWith('doesNotExist', -1, null, client.socketWrapper);
            chai_1.expect(errorCallback).to.have.callCount(0);
            chai_1.expect(services.cache.lastRequestedKey).to.equal('doesNotExist');
            chai_1.expect(services.storage.lastRequestedKey).to.equal('doesNotExist');
        });
        it('handles cache errors', () => {
            services.cache.nextGetWillBeSynchronous = true;
            services.cache.nextOperationWillBeSuccessful = false;
            recordRequest('cacheError', config, services, client.socketWrapper, completeCallback, errorCallback, null);
            chai_1.expect(errorCallback).to.have.been.calledWith(C.RECORD_ACTIONS.RECORD_LOAD_ERROR, 'error while loading cacheError from cache:storageError', 'cacheError', client.socketWrapper);
            chai_1.expect(completeCallback).to.have.callCount(0);
            chai_1.expect(services.logger._log).to.have.been.calledWith(3, C.RECORD_ACTIONS.RECORD_LOAD_ERROR, 'error while loading cacheError from cache:storageError');
            // expect(client.socketWrapper.socket.lastSendMessage).to.equal(
            //   msg('R|E|RECORD_LOAD_ERROR|error while loading cacheError from cache:storageError+'
            // ))
        });
        it('handles storage errors', () => {
            services.cache.nextGetWillBeSynchronous = true;
            services.cache.nextOperationWillBeSuccessful = true;
            services.storage.nextGetWillBeSynchronous = true;
            services.storage.nextOperationWillBeSuccessful = false;
            recordRequest('storageError', config, services, client.socketWrapper, completeCallback, errorCallback, null);
            chai_1.expect(errorCallback).to.have.been.calledWith(C.RECORD_ACTIONS.RECORD_LOAD_ERROR, 'error while loading storageError from storage:storageError', 'storageError', client.socketWrapper);
            chai_1.expect(completeCallback).to.have.callCount(0);
            chai_1.expect(services.logger._log).to.have.been.calledWith(3, C.RECORD_ACTIONS.RECORD_LOAD_ERROR, 'error while loading storageError from storage:storageError');
            // expect(socketWrapper.socket.lastSendMessage).to.equal(msg('R|E|RECORD_LOAD_ERROR|error while loading storageError from storage:storageError+'))
        });
        describe('handles cache timeouts', () => {
            beforeEach(() => {
                config.cacheRetrievalTimeout = 1;
                services.cache.nextGetWillBeSynchronous = false;
                services.cache.nextOperationWillBeSuccessful = true;
            });
            afterEach(() => {
                config.cacheRetrievalTimeout = 10;
            });
            it('sends a CACHE_RETRIEVAL_TIMEOUT message when cache times out', (done) => {
                recordRequest('willTimeoutCache', config, services, client.socketWrapper, completeCallback, errorCallback, null);
                setTimeout(() => {
                    chai_1.expect(errorCallback).to.have.been.calledWith(C.RECORD_ACTIONS.CACHE_RETRIEVAL_TIMEOUT, 'willTimeoutCache', 'willTimeoutCache', client.socketWrapper);
                    chai_1.expect(completeCallback).to.have.callCount(0);
                    // ignores update from cache that may occur afterwards
                    services.cache.triggerLastGetCallback(null, '{ data: "value" }');
                    chai_1.expect(completeCallback).to.have.callCount(0);
                    done();
                }, 1);
            });
        });
        describe('handles storage timeouts', () => {
            beforeEach(() => {
                config.storageRetrievalTimeout = 1;
                services.cache.nextGetWillBeSynchronous = true;
                services.cache.nextOperationWillBeSuccessful = true;
                services.storage.nextGetWillBeSynchronous = false;
                services.storage.nextOperationWillBeSuccessful = true;
            });
            it('sends a STORAGE_RETRIEVAL_TIMEOUT message when storage times out', (done) => {
                recordRequest('willTimeoutStorage', config, services, client.socketWrapper, completeCallback, errorCallback, null);
                setTimeout(() => {
                    chai_1.expect(errorCallback).to.have.been.calledWith(C.RECORD_ACTIONS.STORAGE_RETRIEVAL_TIMEOUT, 'willTimeoutStorage', 'willTimeoutStorage', client.socketWrapper);
                    chai_1.expect(completeCallback).to.have.callCount(0);
                    // ignores update from storage that may occur afterwards
                    services.storage.triggerLastGetCallback(null, '{ data: "value" }');
                    chai_1.expect(completeCallback).to.have.callCount(0);
                    done();
                }, 1);
            });
        });
    });
    describe('excluded records are not put into storage', () => {
        beforeEach(() => {
            services.cache.nextGetWillBeSynchronous = true;
            services.storage.nextGetWillBeSynchronous = true;
            services.storage.delete = sinon_1.spy();
            services.storage.set('dont-save/1', 1, {}, () => { });
        });
        it('returns null when requesting a record that doesn\'t exists in a synchronous cache, and is excluded from storage', (done) => {
            recordRequest('dont-save/1', config, services, client.socketWrapper, completeCallback, errorCallback, null);
            chai_1.expect(completeCallback).to.have.been.calledWith('dont-save/1', -1, null, client.socketWrapper);
            chai_1.expect(errorCallback).to.have.callCount(0);
            chai_1.expect(services.storage.lastRequestedKey).to.equal(null);
            done();
        });
    });
});
//# sourceMappingURL=record-request.spec.js.map