"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
require("mocha");
const chai_1 = require("chai");
const sinon_1 = require("sinon");
const C = require("../constants");
const dependency_initialiser_1 = require("./dependency-initialiser");
const plugin_mock_1 = require("../test/mock/plugin-mock");
const logger_mock_1 = require("../test/mock/logger-mock");
const services = {
    logger: new logger_mock_1.default()
};
describe('dependency-initialiser', () => {
    let dependencyBInitialiser;
    const config = {
        pluginA: new plugin_mock_1.default({}, 'A'),
        pluginB: new plugin_mock_1.default({}, 'B'),
        pluginC: new plugin_mock_1.default({}, 'C'),
        brokenPlugin: {},
        dependencyInitialisationTimeout: 50
    };
    it('throws an error if dependency doesnt implement emitter or has isReady', () => {
        chai_1.expect(() => {
            // tslint:disable-next-line:no-unused-expression
            new dependency_initialiser_1.default({}, config, services, {}, 'brokenPlugin');
        }).to.throw();
        chai_1.expect(services.logger.lastLogEvent).to.equal(C.EVENT.PLUGIN_INITIALIZATION_ERROR);
    });
    it('selects the correct plugin', () => {
        services.logger.lastLogEvent = null;
        dependencyBInitialiser = new dependency_initialiser_1.default({}, config, services, config.pluginB, 'pluginB');
        chai_1.expect(dependencyBInitialiser.getDependency().description).to.equal('B');
        chai_1.expect(services.logger.lastLogEvent).to.equal(null);
    });
    it('notifies when the plugin is ready', (done) => {
        const readySpy = sinon_1.spy();
        dependencyBInitialiser.on('ready', readySpy);
        config.pluginB.setReady();
        setTimeout(() => {
            chai_1.expect(services.logger.lastLogEvent).to.equal(C.EVENT.INFO);
            chai_1.expect(readySpy).to.have.callCount(1);
            done();
        }, 5);
    });
    it('sets deepstream on the plugin if setDeepstream is present', () => {
        const dsMock = { is: 'deepstream' };
        const setDsSpy = config.pluginC.setDeepstream = sinon_1.spy();
        config.pluginC.isReady = true;
        // tslint:disable-next-line:no-unused-expression
        new dependency_initialiser_1.default(dsMock, config, services, config.pluginC, 'pluginC');
        chai_1.expect(setDsSpy).to.have.been.calledWith(dsMock);
    });
    it('allows plugins to become ready after deepstream is set', () => {
        const dsMock = { is: 'deepstream' };
        config.pluginC.deepstream = null;
        config.pluginC.setDeepstream = (deepstream) => {
            config.pluginC.deepstream = deepstream;
            config.pluginC.setReady();
        };
        config.pluginC.isReady = false;
        // tslint:disable-next-line:no-unused-expression
        new dependency_initialiser_1.default(dsMock, config, services, config.pluginC, 'pluginC');
        chai_1.expect(config.pluginC.deepstream).to.equal(dsMock);
    });
});
describe('encounters timeouts and errors during dependency initialisations', () => {
    let dependencyInitialiser;
    const onReady = sinon_1.spy();
    const originalConsoleLog = console.log;
    const config = {
        plugin: new plugin_mock_1.default('A'),
        dependencyInitialisationTimeout: 1
    };
    it('disables console.error', () => {
        Object.defineProperty(console, 'error', {
            value: services.logger.log
        });
    });
    it('creates a depdendency initialiser and doesnt initialise a plugin in time', (done) => {
        dependencyInitialiser = new dependency_initialiser_1.default({}, config, services, config.plugin, 'plugin');
        dependencyInitialiser.on('ready', onReady);
        chai_1.expect(config.plugin.isReady).to.equal(false);
        process.removeAllListeners('uncaughtException');
        process.once('uncaughtException', () => {
            chai_1.expect(services.logger._log).to.have.been.calledWith(3, C.EVENT.PLUGIN_ERROR, 'plugin wasn\'t initialised in time');
            done();
        });
        chai_1.expect(onReady).to.have.callCount(0);
    });
    it.skip('creates another depdendency initialiser with a plugin error', (next) => {
        process.once('uncaughtException', () => {
            chai_1.expect(onReady).to.have.callCount(0);
            chai_1.expect(services.logger._log).to.have.been.calledWith('Error while initialising dependency');
            chai_1.expect(services.logger._log).to.have.been.calledWith('Error while initialising plugin: something went wrong');
            next();
        });
        dependencyInitialiser = new dependency_initialiser_1.default({}, config, services, config.plugin, 'plugin');
        dependencyInitialiser.on('ready', onReady);
        try {
            config.plugin.emit('error', 'something went wrong');
            next('Fail');
        }
        catch (err) { }
    });
    it('enable console.error', () => {
        Object.defineProperty(console, 'error', {
            value: originalConsoleLog
        });
    });
});
//# sourceMappingURL=dependency-initialiser.spec.js.map