"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const constants_1 = require("../constants");
class SubscriptionRegistry {
    /**
     * A generic mechanism to handle subscriptions from sockets to topics.
     * A bit like an event-hub, only that it registers SocketWrappers rather
     * than functions
     */
    constructor(config, services, topic, clusterTopic) {
        this.sockets = new Map();
        this.subscriptions = new Map();
        this.config = config;
        this.services = services;
        this.topic = topic;
        this.clusterTopic = clusterTopic;
        switch (topic) {
            case constants_1.TOPIC.RECORD:
                this.actions = constants_1.RECORD_ACTIONS;
                break;
            case constants_1.TOPIC.EVENT:
                this.actions = constants_1.EVENT_ACTIONS;
                break;
            case constants_1.TOPIC.RPC:
                this.actions = constants_1.RPC_ACTIONS;
                break;
            case constants_1.TOPIC.PRESENCE:
                this.actions = constants_1.PRESENCE_ACTIONS;
                break;
        }
        this.constants = {
            MULTIPLE_SUBSCRIPTIONS: this.actions.MULTIPLE_SUBSCRIPTIONS,
            NOT_SUBSCRIBED: this.actions.NOT_SUBSCRIBED,
            SUBSCRIBE: this.actions.SUBSCRIBE,
            UNSUBSCRIBE: this.actions.UNSUBSCRIBE,
        };
        this.onSocketClose = this.onSocketClose.bind(this);
        this.setupRemoteComponents(clusterTopic);
    }
    whenReady(callback) {
        this.clusterSubscriptions.whenReady(callback);
    }
    /**
     * Setup all the remote components and actions required to deal with the subscription
     * via the cluster.
     */
    setupRemoteComponents(clusterTopic) {
        this.clusterSubscriptions = this.services.message.getStateRegistry(clusterTopic);
    }
    /**
     * Return all the servers that have this subscription.
     */
    getAllServers(subscriptionName) {
        return this.clusterSubscriptions.getAllServers(subscriptionName);
    }
    /**
     * Return all the servers that have this subscription excluding the current
     * server name
     */
    getAllRemoteServers(subscriptionName) {
        const serverNames = this.clusterSubscriptions.getAllServers(subscriptionName);
        const localServerIndex = serverNames.indexOf(this.config.serverName);
        if (localServerIndex > -1) {
            serverNames.splice(serverNames.indexOf(this.config.serverName), 1);
        }
        return serverNames;
    }
    /**
     * Returns a list of all the topic this registry
     * currently has subscribers for
     */
    getNames() {
        return this.clusterSubscriptions.getAll();
    }
    /**
     * Returns a map of all the topic this registry
     * currently has subscribers for
     */
    getNamesMap() {
        return this.clusterSubscriptions.getAllMap();
    }
    /**
     * Returns true if the subscription exists somewhere
     * in the cluster
     */
    hasName(subscriptionName) {
        return this.clusterSubscriptions.has(subscriptionName);
    }
    /**
    * This method allows you to customise the SubscriptionRegistry so that it can send
    * custom events and ack messages back.
    * For example, when using the ACTIONS.LISTEN, you would override SUBSCRIBE with
    * ACTIONS.SUBSCRIBE and UNSUBSCRIBE with UNSUBSCRIBE
    */
    setAction(name, value) {
        this.constants[name.toUpperCase()] = value;
    }
    /**
     * Enqueues a message string to be broadcast to all subscribers. Broadcasts will potentially
     * be reordered in relation to *other* subscription names, but never in relation to the same
     * subscription name. Each broadcast is given 'broadcastTimeout' ms to coalesce into one big
     * broadcast.
     */
    sendToSubscribers(name, message, noDelay, senderSocket, isRemote = false) {
        if (senderSocket && !isRemote) {
            this.services.message.send(this.clusterTopic, message);
        }
        const subscription = this.subscriptions.get(name);
        if (!subscription) {
            return;
        }
        const subscribers = subscription.sockets;
        const first = subscribers.values().next().value;
        const msg = first.getMessage(message);
        for (const socket of subscribers) {
            if (socket === senderSocket) {
                continue;
            }
            socket.sendBinaryMessage(msg, true);
        }
    }
    /**
     * Adds a SocketWrapper as a subscriber to a topic
     */
    subscribe(message, socket, silent) {
        const name = message.name;
        const subscription = this.subscriptions.get(name) || {
            name,
            sockets: new Set()
        };
        if (subscription.sockets.size === 0) {
            this.subscriptions.set(name, subscription);
        }
        else if (subscription.sockets.has(socket)) {
            const msg = `repeat subscription to "${name}" by ${socket.user}`;
            this.services.logger.warn(constants_1.EVENT_ACTIONS[this.constants.MULTIPLE_SUBSCRIPTIONS], msg);
            socket.sendMessage({
                topic: this.topic,
                action: this.constants.MULTIPLE_SUBSCRIPTIONS,
                originalAction: message.action,
                name
            });
            return;
        }
        subscription.sockets.add(socket);
        this.addSocket(subscription, socket);
        if (!silent) {
            const logMsg = `for ${constants_1.TOPIC[this.topic]}:${name} by ${socket.user}`;
            this.services.logger.debug(this.actions[this.constants.SUBSCRIBE], logMsg);
            socket.sendAckMessage(message);
        }
    }
    /**
     * Removes a SocketWrapper from the list of subscriptions for a topic
     */
    unsubscribe(message, socket, silent) {
        const name = message.name;
        const subscription = this.subscriptions.get(name);
        if (!subscription || !subscription.sockets.delete(socket)) {
            if (!silent) {
                const msg = `${socket.user} is not subscribed to ${name}`;
                this.services.logger.warn(this.actions[this.constants.NOT_SUBSCRIBED], msg);
                socket.sendMessage({
                    topic: this.topic,
                    action: this.constants.NOT_SUBSCRIBED,
                    originalAction: message.action,
                    name
                });
            }
            return;
        }
        this.removeSocket(subscription, socket);
        if (!silent) {
            const logMsg = `for ${this.topic}:${name} by ${socket.user}`;
            this.services.logger.debug(this.actions[this.constants.UNSUBSCRIBE], logMsg);
            socket.sendAckMessage(message);
        }
    }
    /**
     * Returns an array of SocketWrappers that are subscribed
     * to <name> or null if there are no subscribers
     */
    getLocalSubscribers(name) {
        const subscription = this.subscriptions.get(name);
        return subscription ? subscription.sockets : new Set();
    }
    /**
     * Returns true if there are SocketWrappers that
     * are subscribed to <name> or false if there
     * aren't any subscribers
     */
    hasLocalSubscribers(name) {
        return this.subscriptions.has(name);
    }
    /**
     * Allows to set a subscriptionListener after the class had been instantiated
     */
    setSubscriptionListener(listener) {
        this.subscriptionListener = listener;
        this.clusterSubscriptions.on('add', listener.onFirstSubscriptionMade.bind(listener));
        this.clusterSubscriptions.on('remove', listener.onLastSubscriptionRemoved.bind(listener));
    }
    addSocket(subscription, socket) {
        const subscriptions = this.sockets.get(socket) || new Set();
        if (subscriptions.size === 0) {
            this.sockets.set(socket, subscriptions);
            socket.onClose(this.onSocketClose);
        }
        subscriptions.add(subscription);
        this.clusterSubscriptions.add(subscription.name);
        if (this.subscriptionListener) {
            this.subscriptionListener.onSubscriptionMade(subscription.name, socket);
        }
    }
    removeSocket(subscription, socket) {
        if (subscription.sockets.size === 0) {
            this.subscriptions.delete(subscription.name);
            socket.removeOnClose(this.onSocketClose);
        }
        if (this.subscriptionListener) {
            this.subscriptionListener.onSubscriptionRemoved(subscription.name, socket);
        }
        this.clusterSubscriptions.remove(subscription.name);
        const subscriptions = this.sockets.get(socket);
        if (subscriptions) {
            subscriptions.delete(subscription);
        }
        else {
            // log error
        }
    }
    /**
    * Called whenever a socket closes to remove all of its subscriptions
    */
    onSocketClose(socket) {
        const subscriptions = this.sockets.get(socket);
        if (!subscriptions) {
            this.services.logger.error(constants_1.EVENT_ACTIONS[this.constants.NOT_SUBSCRIBED], 'A socket has an illegal registered close callback');
            return;
        }
        for (const subscription of subscriptions) {
            subscription.sockets.delete(socket);
            this.removeSocket(subscription, socket);
        }
    }
}
exports.default = SubscriptionRegistry;
//# sourceMappingURL=subscription-registry.js.map