"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
require("mocha");
const sinon_1 = require("sinon");
const chai_1 = require("chai");
const events_1 = require("events");
const utils = require('./utils');
describe('utils', () => {
    it('receives a different value everytime getUid is called', () => {
        const uidA = utils.getUid();
        const uidB = utils.getUid();
        const uidC = utils.getUid();
        chai_1.expect(uidA).not.to.equal(uidB);
        chai_1.expect(uidB).not.to.equal(uidC);
        chai_1.expect(uidA).not.to.equal(uidC);
    });
    it('combines multiple events into one', () => {
        const emitters = [
            new events_1.EventEmitter(),
            new events_1.EventEmitter(),
            new events_1.EventEmitter()
        ];
        const callback = sinon_1.spy();
        utils.combineEvents(emitters, 'someEvent', callback);
        chai_1.expect(callback).to.have.callCount(0);
        emitters[0].emit('someEvent');
        chai_1.expect(callback).to.have.callCount(0);
        emitters[1].emit('someEvent');
        emitters[2].emit('someEvent');
        chai_1.expect(callback).to.have.callCount(1);
    });
    it('reverses maps', () => {
        const user = {
            firstname: 'Wolfram',
            lastname: 'Hempel'
        };
        chai_1.expect(utils.reverseMap(user)).to.deep.equal({
            Wolfram: 'firstname',
            Hempel: 'lastname'
        });
    });
    describe('isOfType', () => {
        it('checks basic types', () => {
            chai_1.expect(utils.isOfType('bla', 'string')).to.equal(true);
            chai_1.expect(utils.isOfType(42, 'string')).to.equal(false);
            chai_1.expect(utils.isOfType(42, 'number')).to.equal(true);
            chai_1.expect(utils.isOfType(true, 'number')).to.equal(false);
            chai_1.expect(utils.isOfType(true, 'boolean')).to.equal(true);
            chai_1.expect(utils.isOfType({}, 'object')).to.equal(true);
            chai_1.expect(utils.isOfType(null, 'null')).to.equal(true);
            chai_1.expect(utils.isOfType(null, 'object')).to.equal(false);
            chai_1.expect(utils.isOfType([], 'object')).to.equal(true);
        });
        it('checks urls', () => {
            chai_1.expect(utils.isOfType('bla', 'url')).to.equal(false);
            chai_1.expect(utils.isOfType('bla:22', 'url')).to.equal(true);
            chai_1.expect(utils.isOfType('https://deepstream.io/', 'url')).to.equal(true);
        });
        it('checks arrays', () => {
            chai_1.expect(utils.isOfType([], 'array')).to.equal(true);
            chai_1.expect(utils.isOfType({}, 'array')).to.equal(false);
        });
    });
    describe('validateMap', () => {
        function _map() {
            return {
                'a-string': 'bla',
                'a number': 42,
                'an array': ['yup']
            };
        }
        function _schema() {
            return {
                'a-string': 'string',
                'a number': 'number',
                'an array': 'array'
            };
        }
        it('validates basic maps', () => {
            const map = _map();
            const schema = _schema();
            chai_1.expect(utils.validateMap(map, false, schema)).to.equal(true);
        });
        it('fails validating an incorrect map', () => {
            const map = _map();
            const schema = _schema();
            schema['an array'] = 'number';
            const returnValue = utils.validateMap(map, false, schema);
            chai_1.expect(returnValue instanceof Error).to.equal(true);
        });
        it('fails validating an incomplete map', () => {
            const map = _map();
            const schema = _schema();
            delete map['an array'];
            const returnValue = utils.validateMap(map, false, schema);
            chai_1.expect(returnValue instanceof Error).to.equal(true);
        });
        it('throws errors', () => {
            const map = _map();
            const schema = _schema();
            schema['an array'] = 'number';
            chai_1.expect(() => {
                utils.validateMap(map, true, schema);
            }).to.throw();
        });
    });
    describe('merges recoursively', () => {
        it('merges two simple objects', () => {
            const objA = {
                firstname: 'Homer',
                lastname: 'Simpson'
            };
            const objB = {
                firstname: 'Marge'
            };
            chai_1.expect(utils.merge(objA, objB)).to.deep.equal({
                firstname: 'Marge',
                lastname: 'Simpson'
            });
        });
        it('merges two nested objects', () => {
            const objA = {
                firstname: 'Homer',
                lastname: 'Simpson',
                children: {
                    Bart: {
                        lastname: 'Simpson'
                    }
                }
            };
            const objB = {
                firstname: 'Marge',
                children: {
                    Bart: {
                        firstname: 'Bart'
                    }
                }
            };
            chai_1.expect(utils.merge(objA, objB)).to.deep.equal({
                firstname: 'Marge',
                lastname: 'Simpson',
                children: {
                    Bart: {
                        firstname: 'Bart',
                        lastname: 'Simpson'
                    }
                }
            });
        });
        it('merges multiple objects ', () => {
            const objA = {
                pets: {
                    birds: ['parrot', 'dove']
                }
            };
            const objB = {
                jobs: {
                    hunter: false
                }
            };
            const objC = {
                firstname: 'Egon'
            };
            chai_1.expect(utils.merge(objA, objB, {}, objC)).to.deep.equal({
                pets: {
                    birds: ['parrot', 'dove']
                },
                jobs: {
                    hunter: false
                },
                firstname: 'Egon'
            });
        });
        it('handles null and undefined values', () => {
            const objA = {
                pets: {
                    dog: 1,
                    cat: 2,
                    ape: 3
                }
            };
            const objB = {
                pets: {
                    cat: null,
                    ape: undefined,
                    zebra: 9
                }
            };
            chai_1.expect(utils.merge(objA, objB)).to.deep.equal({
                pets: {
                    dog: 1,
                    cat: null,
                    ape: 3,
                    zebra: 9
                }
            });
        });
    });
});
//# sourceMappingURL=utils.spec.js.map